PrecacheSound("vo/null.mp3");

::ProvideCustomWeapon <- function(player, customWeaponClass)
{
    local weaponEnt = CreateByClassname(customWeaponClass.baseclass);
    SetPropInt(weaponEnt, "m_AttributeManager.m_Item.m_iItemDefinitionIndex", customWeaponClass.baseid);
    SetPropBool(weaponEnt, "m_AttributeManager.m_Item.m_bInitialized", true);
    SetPropBool(weaponEnt, "m_bValidatedAttachedEntity", true);
	SetPropBool(weaponEnt, "m_bClientSideAnimation", true);
    weaponEnt.SetTeam(player.GetTeam());
    weaponEnt.SetOwner(player);
	SetPropEntity(weaponEnt, "m_hOwnerEntity", player);
    weaponEnt.DispatchSpawn();
    player.Weapon_Equip(weaponEnt);

    local instance = customWeaponClass(player, weaponEnt);
    instance.ApplyWeapon(player);
    weaponEnt.ValidateScriptScope();
    weaponEnt.GetScriptScope().customWeapon <- instance;
    return weaponEnt;
}

::CustomWeaponBase <- class
{
    //Init data
    player = null;
    charScript = null;
    weapon = null;

    baseclass = null;
    baseid = null;
    hideworldmodel = null;
    clip = null;
    maxAmmo = null;
    //independentReserveAmmo = false; //todo

    viewmodel = null;
    worldmodel = null;
    playermodel = null;
    armsmodel = null;
    extravm = null; //[];
    extrawm = null; //[];

    vmIndex = 0;
    wmIndex = 0;
    evmIndices = null; //[];
    ewmIndices = null; //[];

    mute = false;

    //Dynamic data
    playerInTaunt = false;
    weaponActiveLastTick = null;
    lastFireTS = 0;
    extramodelEnts = null; //[];

    //Precaching
    needsPrecaching = true;
    function Precache() { }

    constructor(player, weapon)
    {
        local myClass = getclass();
        if (myClass.needsPrecaching)
        {
            myClass.needsPrecaching = false;
            if (playermodel) PrecacheModel(playermodel);
            myClass.vmIndex = viewmodel ? PrecacheModel(viewmodel) : 0;
            myClass.wmIndex = worldmodel ? PrecacheModel(worldmodel) : 0;
            if (worldmodel)
            {
                if (!myClass.extrawm)
                    myClass.extrawm = [];
                //myClass.extrawm.push(worldmodel);
            }
            if (armsmodel)
            {
                if (!myClass.extravm)
                    myClass.extravm = [];
                myClass.extravm.push(armsmodel);
            }
            if (myClass.extravm)
            {
                myClass.evmIndices = [];
                foreach(model in myClass.extravm)
                    myClass.evmIndices.push(PrecacheModel(model));
            }
            if (myClass.extrawm)
            {
                myClass.ewmIndices = [];
                foreach(model in myClass.extrawm)
                    myClass.ewmIndices.push(PrecacheModel(model));
            }
            Precache();
        }
        vmIndex = myClass.vmIndex;
        wmIndex = myClass.wmIndex;
        evmIndices = myClass.evmIndices;
        ewmIndices = myClass.ewmIndices;
        extramodelEnts = [];

        this.player = player;
        this.weapon = weapon;
        this.charScript = "GetCustomCharacter" in player ? player.GetCustomCharacter() : null;
        OnSelfEvent("clear_custom_character", 0, OnCleanupInternal);

        AddTimer(-1, function()
        {
            //Weapon Switching
            local activeWeapon = player.GetActiveWeapon();

            if (weapon != activeWeapon && weapon == weaponActiveLastTick)
            {
                foreach (extramodelEnt in extramodelEnts)
                    KillIfValid(extramodelEnt);
                OnSwitchOff();
            }
            if (weapon == activeWeapon && weapon != weaponActiveLastTick)
                DoOnSwitch();

            weaponActiveLastTick = activeWeapon;

            if (activeWeapon != weapon)
                return;

            //Taunting
            local inCond = player.InCond(TF_COND_TAUNTING);
            if (inCond && !playerInTaunt)
            {
                playerInTaunt = true;
            }
            else if (!inCond && playerInTaunt)
            {
                playerInTaunt = false;
                SetPropBool(weapon, "m_bBeingRepurposedForTaunt", true);
            }

            //Firing
            local newFireTS = GetPropFloat(weapon, "m_flLastFireTime");
            if (newFireTS > lastFireTS)
            {
                lastFireTS = newFireTS;
                OnFire();
            }
        });
    }

    function OnCleanupInternal()
    {
        foreach (extramodelEnt in extramodelEnts)
            KillIfValid(extramodelEnt);
        OnCleanup();
        player = null;
        weapon = null;
    }

    function IsActive()
    {
        //todo why player null
        return player && player.GetActiveWeapon() == weapon;
    }

    function SetModels()
    {
        if (viewmodel)
        {
            weapon.SetModel(viewmodel);
            weapon.SetCustomViewModelModelIndex(vmIndex);
            SetPropInt(weapon, "m_nModelIndex", vmIndex);
            SetPropInt(weapon, "m_iViewModelIndex", vmIndex);
        }
        if (hideworldmodel == true || (wmIndex && hideworldmodel != false))
        {
            //if (player.GetMaxHealth() < 400) //todo kludge
            if (false)
            {
                weapon.KeyValueFromInt("effects", 161);
                weapon.DisableDraw();
                SetPropInt(weapon, "m_nRenderMode", 1);
                SetEntityColor(weapon, [0, 0, 0, 0]);
                SetPropBool(weapon, "m_bBeingRepurposedForTaunt", true);
            }
            else
            {
                weapon.KeyValueFromInt("effects", 161);
                SetPropBool(weapon, "m_bBeingRepurposedForTaunt", true);

                for (local i = 0; i < 4; i++)
                    SetPropIntArray(weapon, "m_nModelIndexOverrides", wmIndex, i)
            }

            /*weapon.KeyValueFromInt("effects", 161);
            weapon.DisableDraw();
            SetPropInt(weapon, "m_nRenderMode", 1);
            SetEntityColor(weapon, [0, 0, 0, 0]);
            SetPropBool(weapon, "m_bBeingRepurposedForTaunt", true);*/

            //weapon.SetModelSimple(worldmodel);
            //SetPropInt(weapon, "m_nModelIndex", wmIndex);
            //SetPropInt(weapon, "m_iViewModelIndex", wmIndex);
            //for (local i = 0; i < 4; i++)
             //SetPropIntArray(weapon, "m_nModelIndexOverrides", wmIndex, i);
        }
    }

    function ApplyWeapon(player)
    {
        this.player = player;
        SetModels();
        if (clip != null)
            weapon.SetClip1(clip);
        if (maxAmmo != null)
        {
            local ammoType = weapon.GetPrimaryAmmoType();
            if (ammoType >= 0)
            {
                SetPropIntArray(player, "m_iAmmo", maxAmmo, ammoType);

                local attribute = ammoType == TF_AMMO_PRIMARY
                    ? "hidden primary max ammo bonus"
                    : "hidden secondary max ammo penalty";
                local maxAmmoBuff = maxAmmo / TF_CLASS_AMMO[player.GetPlayerClass()][ammoType].tofloat();
                weapon.AddAttribute(attribute, maxAmmoBuff, -1);
            }
        }
        OnEquip();
    }

    function DoOnSwitch()
    {
        if (evmIndices) foreach (modelIndex in evmIndices)
        {
            local modelEnt = SpawnEntityFromTable("tf_wearable_vm", {});
            modelEnt.Teleport(true, player.GetOrigin(), true, player.GetAbsAngles(), false, Vector());
            SetPropInt(modelEnt, "m_nModelIndex", modelIndex);
            player.EquipWearableViewModel(modelEnt);
            extramodelEnts.push(modelEnt);
        }
        if (ewmIndices) foreach (modelIndex in ewmIndices)
        {
            local modelEnt = SpawnEntityFromTable("tf_wearable", { });
            modelEnt.Teleport(true, player.GetOrigin(), true, player.GetAbsAngles(), false, Vector());
            SetPropInt(modelEnt, "m_nModelIndex", modelIndex);
            SetPropBool(modelEnt, "m_bValidatedAttachedEntity", true);
            SetPropBool(modelEnt, "m_AttributeManager.m_Item.m_bInitialized", true);
            SetPropEntity(modelEnt, "m_hOwnerEntity", player);
            EntFireByHandle(modelEnt, "SetParent", "!activator", 0, player, player);
            extramodelEnts.push(modelEnt);
        }
        if (playermodel)
            player.SetCustomModelWithClassAnimations(playermodel);
        if (mute)
        {
            AddTimer(-1, function()
            {
                if (!IsActive())
                    return TIMER_DELETE;
                EmitSoundEx({
                    sound_name = "vo/null.mp3",
                    entity = player,
                    channel = CHAN_WEAPON,
                    sound_level = 150,
                    volume = 1,
                });
            });
        }
        OnSwitch();
    }

    function OnEquip() { }
    function OnSwitch() { }
    function OnSwitchOff() { }
    function OnFire() { }
    function OnCleanup() { }
}