PrecacheModel("models/Weapons/w_grenade.mdl"); //todo do we still need this?

::projectileScripts <- {};

class ::ProjectileBase
{
    damage = null;
    model = null;
    offset = null;
    rotOffset = null;
    speed = null;

    shooter = null;
    weapon = null;
    hasExpired = false;

    constructor()
    {
    }

    function Shoot(shooter, weapon)
    {
        this.shooter = shooter;
        this.weapon = weapon;

        local position = shooter.EyePosition();
        local rotation = shooter.EyeAngles();

        if (rotOffset) rotation += rotOffset;
        if (offset) position += RotatePosition(Vector(), rotation, offset);

        local projectile = Spawn(position, rotation);
        SetDestroyCallback(projectile, OnDestroy.bindenv(this));
        return projectile;
    }

    function ShootAbs(shooter, weapon, location, rotation, velocity)
    {
        this.shooter = shooter;
        this.weapon = weapon;

        local projectile = Spawn(position, rotation, velocity);
        SetDestroyCallback(projectile, OnDestroy.bindenv(this));
        return projectile;
    }

    function Spawn(position, rotation)
    {
    }

    function OnDetonate()
    {
        printl("OnDetonate: "+hasExpired)
    }

    function OnDestroy()
    {
    }
}

class ::ProjectileGrenade extends ProjectileBase
{
    fusetime = null;
    radius = null;
    shouldExplode = null;
    explodeOnTouch = true;

    weapon_mimic = null;

    function Spawn(position, rotation)
    {
        local rf = RandomFloat(0, 1);
        weapon_mimic = SpawnEntityFromTable("tf_point_weapon_mimic", {
            WeaponType = 1,
            origin = position,
            angles = rotation,
            SpeedMin = speed,
            SpeedMax = speed,
            Damage = damage,
            SplashRadius = rf,
            ModelOverride = model,
            ModelScale = 1,
        });
        weapon_mimic.AcceptInput("FireOnce", "", shooter, shooter);

        for (local projectile = null; projectile = FindByClassnameWithin(projectile, "tf_projectile_pipe", position, 100);)
            if (abs(rf - GetPropFloat(projectile, "m_DmgRadius")) <= Epsilon)
                return SetupProjectile(projectile);

        return null;
    }

    function SetupProjectile(projectile)
    {
        projectileScripts[projectile] <- this;

        SetPropFloat(projectile, "m_DmgRadius", radius);
        SetPropEntity(projectile, "m_hThrower", shooter);
        projectile.SetTeam(shooter.GetTeam());
        projectile.AddEFlags(EFL_NO_THINK_FUNCTION);

        AddTimer(-1, function()
        {
            local takeDamage = GetPropInt(projectile, "m_takedamage"); //todo magic values
            if (takeDamage == -1)
            {
                KillIfValid(weapon_mimic);
                delete projectileScripts[projectile];
            }
            else if (takeDamage == 0)
            {
                OnDetonate();
            }
            else if ((explodeOnTouch && GetPropBool(projectile, "m_bTouched")))
            {
                OnDetonate();
                SetPropBool(projectile, "m_bTouched", false);
                projectile.SetSize(Vector(-6000, -6000, -6000), Vector(6000, 6000, 6000));
                SetPropInt(projectile, "m_takedamage", -1);
                OnNextTick(KillIfValid, projectile);
            }
        });

        if (fusetime != null)
            RunWithDelay(fusetime, function()
            {
                hasExpired = true;
                OnDetonate();
                SetPropBool(projectile, "m_bTouched", false);
                projectile.SetSize(Vector(-6000, -6000, -6000), Vector(6000, 6000, 6000));
                projectile.RemoveEFlags(EFL_NO_THINK_FUNCTION);
            });

        return projectile;
    }
}

class ::ProjectileCleaver extends ProjectileBase
{
    fusetime = null;

    function Spawn(position, rotation, velocity)
    {
        local projectile = CreateByClassname("tf_projectile_throwable");
        projectile.AddFlag(2097152);
        projectile.KeyValueFromString("classname", "prop_physics");
        projectile.SetOwner(shooter);
        SetPropEntity(projectile, "m_hThrower", shooter);
        SetPropFloat(projectile, "m_flDamage", damage);
        projectile.Teleport(true, position, true, rotation, false, velocity);
        Entities.DispatchSpawn(projectile);
        projectile.SetModelSimple(model);
        projectile.SetTeam(shooter.GetTeam());
        projectile.SetPhysVelocity(velocity);
        projectile.AddEFlags(EFL_NO_THINK_FUNCTION);
        projectile.ValidateScriptScope();
        projectile.GetScriptScope().script <- this;

        if (fusetime)
        {
            RunWithDelay(fusetime, OnExpire);
            EntFireByHandle(projectile, "Kill", null, fusetime, null, null);
        }

        return projectile;
    }
}

return;

::ProjectileInfo <- class
{
    projectiles = {};

    damageDirect = null;
    blastDamage = null;
    blastRadius = null;
    damageFalloff = null;

    damageInflictor = null;
    damageType = null;

    model = null;
    speed = null;
    offset = null;
    rotOffset = null;
    gravity = null;

    shooter = null;
    weapon = null;
    rotation = null;
    position = null;

    projectile = null;
    projectileEntIndex = null;
    size = null;
    self = null;

    shouldHitTeammates = null;
    onHit = null;

    constructor()
    {
        projectiles[this] <- 1;
    }

    function Build(projParams)
    {
        shooter = safeget(projParams, "shooter", null);
        weapon = safeget(projParams, "weapon", null);
        rotation = safeget(projParams, "rotation", shooter.EyeAngles());
        position = safeget(projParams, "position", shooter.EyePosition());

        if (offset)
            position = position + RotatePosition(Vector(), rotation, offset);
        if (rotOffset)
            rotation = rotation + rotOffset;

        local velocity = safeget(projParams, "velocity", rotation.Forward() * speed);
        local team = safeget(projParams, "team", shooter ? shooter.GetTeam() : 0);

        projectile = CreateByClassname("tf_projectile_throwable");
        projectile.SetAbsOrigin(position);
        projectile.SetAbsAngles(rotation);
        projectile.SetPhysVelocity(velocity);
        projectile.DispatchSpawn();
        projectile.SetModelSimple(model);
        projectile.SetTeam(team);
        projectileEntIndex = projectile.entindex();
        self = main_script_entity;

        if (gravity)
        {
            projectile.SetMoveType(MOVETYPE_FLYGRAVITY, MOVECOLLIDE_DEFAULT);
            projectile.SetGravity(gravity);
            projectile.SetAbsVelocity(velocity);
            //projectile.SetAngularVelocity(RandomFloat(-1000, 1000), RandomFloat(-1000, 1000), RandomFloat(-1000, 1000));
        }
        else
            projectile.SetPhysVelocity(velocity);

        if (!size)
        {
            local delta = projectile.GetBoundingMaxs() - projectile.GetBoundingMins();
            size = max(delta.x, delta.y);
        }
        size = Vector(size, size, size);
        projectile.SetSize(size * -0.5, size * 0.5);

        AddTimer(-1, TickProjectileCollision);
        return projectile;
    }

    function TickProjectileCollision()
    {
        if (projectile.IsValid() && !GetPropBool(projectile, "m_bTouched"))
            return;

        delete projectiles[this];
        projectile = EntIndexToHScript(projectileEntIndex);
        if (!projectile)
            return TIMER_DELETE;

        OnDelete();
    }

    function OnDelete()
    {
        local projectilePos = projectile.GetCenter();
        //If direct hits only
        local directHitTarget = null;
        if (damageDirect)
        {
            local traceHull = {
                start = projectilePos,
                end = projectilePos,
                hullmin = size * -0.5,
                hullmax = size * 0.5
            }
            TraceHull(traceHull);
            directHitTarget = traceHull.hit ? traceHull.enthit : null;
            if (directHitTarget)
            {
                //if (!shouldHitTeammates || (directHitTarget.GetTeam() == shooter.GetTeam()))
                //{
                    OnTouchDirect(directHitTarget);
                //}
            }
        }
        if (blastDamage)
        {
            local explosion = SpawnEntityFromTable("env_explosion",
            {
                iMagnitude = blastDamage,
                iRadiusOverride = blastRadius,
                origin = projectilePos,
                teamnum = shooter.GetTeam(),
                spawnflags = 6740
            });
            explosion.SetOwner(shooter);
            SetPropEntity(explosion, "m_hEntityIgnore", directHitTarget);
            explosion.AcceptInput("Explode", null, shooter, shooter);
        }
        projectile.SetAbsOrigin(Vector(-99999, -99999, -99999));
        KillIfValid(projectile);
        return TIMER_DELETE;
    }

    function OnTouchDirect(target)
    {
        target.TakeDamageEx(
            shooter,
            shooter,
            shooter.GetActiveWeapon(),
            Vector(),
            Vector(),
            damageDirect,
            damageType);
    }
}

OnGameEvent("OnTakeDamage", 0, function(victim, params)
{
    if (params.inflictor && params.inflictor.GetClassname() == "env_explosion")
    {
        //params.damage_type = DMG_SLOWBURN; //todo
        params.attacker = params.inflictor.GetOwner();
    }
});