::AbilityEyeLasers <- class extends BaseAbility
{
    //Custom character definition
    laser_viewmodel = "models/vsh/player/items/vsh_laser_effect.mdl";
    maxCooldown = 30;
    warmupDuration = 3;
    firingDuration = 5;

    //Dynamic data
    custom_dmg_eye_lasers = null;
    warmupEndTS = 0;
    firingEndTS = 0;
    laser_target_ent = null;
    laser_target_helper_ent = null;

    function Precache()
    {
        PrecacheModel(laser_viewmodel);
    }

    function ApplyAbility()
    {
        base.ApplyAbility();
        custom_dmg_eye_lasers = SpawnEntityFromTable("point_template", { classname = "hale_eyelasers" });
        nextCooldownTS = Time() + maxCooldown;
        AddTimer(-1, TickAbility);

        laser_target_ent = SpawnEntityFromTable("prop_dynamic",
        {
            targetname = "laser_target",
            model = "models/empty.mdl"
        });

        EntFireByHandle(SpawnEntityFromTable("prop_dynamic",
        {
            targetname = "laser_target_camera",
            model = "models/empty.mdl"
        }), "SetParent", "!activator", 0, player, player);

        laser_target_helper_ent = SpawnEntityFromTable("prop_dynamic",
        {
            targetname = "laser_target_helper",
            model = "models/empty.mdl"
        });
    }

    function OnCooldownEnd()
    {
        charScript.EmitSoundBankSimple("laser_ready_sfx", {
            filter_type = RECIPIENT_FILTER_GLOBAL
        });
        charScript.EmitSoundBankSimple("laser_warning_sfx", {
            filter_type = RECIPIENT_FILTER_GLOBAL
        });
        RunWithDelay(1, function()
        {
            charScript.EmitVoiceLine("laser_ready", {
                channel = CHAN_VSH_PUNCH
            });
        });
    }

    function CheckInput()
    {
        return player.IsCallingForMedic();
    }

    function IsWarmup()
    {
        return warmupEndTS >= Time();
    }

    function IsFiring()
    {
        return firingEndTS >= Time();
    }

    function CanPeform()
    {
        return base.CanPerform() && !IsWarmup() && !IsFiring();
    }

    function TickAbility()
    {
        if (CanPeform())
            WarmUp();
    }

    function WarmUp()
    {
        warmupEndTS = Time() + warmupDuration + 0.2;
        local firingEndTSLocal = warmupEndTS + firingDuration + 0.5;

        charScript.DisableAttack(firingDuration + 0.5);
        player.AddCustomAttribute("move speed penalty", 0.75, warmupDuration + 0.5);
        player.PlayViewModelAnim("vsh_laser_prepare");

        charScript.EmitVoiceLine("laser_firing", {
            channel = CHAN_VSH_PUNCH
        });
        charScript.EmitSoundBankSimple("laser_fire_sfx", {
            filter_type = RECIPIENT_FILTER_GLOBAL
        });

        RunWithDelay(warmupDuration, StartLaserAttack);
    }

    function StartLaserAttack()
    {
        firingEndTS = Time() + firingDuration;

        AddTimer(0.1, TickLaserAttack);
        RunWithDelay(firingDuration, FinishLaserAttack);

        player.AddCustomAttribute("move speed penalty", 0.5, firingDuration + 0.25);

        //AoE sparks around Mecha Hale
        local particle = SpawnEntityFromTable("info_particle_system", {
            effect_name = "vsh_tesla_shield",
            start_active = 1
        });
        EntFireByHandle(particle, "SetParent", "!activator", 0, player, player);
        EntFireByHandle(particle, "SetParentAttachment", "vsh_tesla_shield", 0, null, null);
        EntFireByHandle(particle, "Kill", "", firingDuration, null, null);

        //Eye laser beam (third person)
        particle = SpawnEntityFromTable("info_particle_system", {
            effect_name = "vsh_laser_tp",
            cpoint1 = "laser_target",
            cpoint2 = "laser_target_camera",
            cpoint3 = "laser_target_helper",
            start_active = 1
        });
        EntFireByHandle(particle, "SetParent", "!activator", 0, player, player);
        EntFireByHandle(particle, "SetParentAttachment", "eyes", 0, null, null);
        EntFireByHandle(particle, "Kill", "", firingDuration, null, null);

        //Eye laser beam (first person)
        PrecacheModel("models/vsh/player/items/vsh_laser_effect.mdl");
        local laserVM = player.CreateViewModel("models/vsh/player/items/vsh_laser_effect.mdl");
        EntFireByHandle(laserVM, "Kill", "", firingDuration, null, null);

        player.PlayViewModelAnim("c_heavy_idle");
    }

    function FinishLaserAttack()
    {
        local viewmodel = GetPropEntity(player, "m_hViewModel");
        if (viewmodel)
            viewmodel.EnableDraw();
        player.PlayViewModelAnim("vsh_show_arms");
        ConsumeCooldown();
    }

    function OnCleanup()
    {
        local viewmodel = GetPropEntity(player, "m_hViewModel");
        if (viewmodel)
            viewmodel.EnableDraw();
        KillIfValid(laser_target_ent);
        KillIfValid(laser_target_helper_ent);
    }

    function TickLaserAttack()
    {
        if (!IsFiring())
            return TIMER_DELETE;

        local forward = player.EyeAngles().Forward();
        local eyePos = player.EyePosition();

        //Aim assist
        local aimAssistTarget = null;

        foreach(target in GetAlivePlayers(player.GetEnemyTeam()))
        {
            if (target == player)
                continue;
            local deltaVector = target.GetCenter() - eyePos;
            local distance = deltaVector.Norm();
            if (distance > 1800)
                continue;
            local dot = forward.Dot(deltaVector);
            if (dot > (distance < 400 ? 0.98 : distance < 1000 ? 0.992 : distance < 1400 ? 0.997 : 0.999))
            {
                local fraction = TraceLine(eyePos, target.GetCenter(), null);
                if (fraction > 0.99)
                {
                    aimAssistTarget = target;
                    break;
                }
            }
        }

        local trace = {
            start = eyePos,
            end = aimAssistTarget ? aimAssistTarget.GetCenter() : eyePos + forward * 1800,
            mask = CONTENTS_SOLID|CONTENTS_MOVEABLE|CONTENTS_PLAYERCLIP|CONTENTS_WINDOW|CONTENTS_GRATE|CONTENTS_MONSTER,
            ignore = player
        };
        TraceLineEx(trace);

        local vecHelper = trace.endpos - trace.startpos;
        vecHelper.Norm();

        local endpos = trace.endpos - vecHelper;

        local frame = NetProps.GetPropInt(laser_target_ent, "m_ubInterpolationFrame");
        laser_target_ent.SetAbsOrigin(endpos);
        NetProps.SetPropInt(laser_target_ent, "m_ubInterpolationFrame", frame);

        vecHelper.z = (trace.fraction == 1.0) ? 1 : 0;

        frame = NetProps.GetPropInt(laser_target_helper_ent, "m_ubInterpolationFrame");
        laser_target_helper_ent.SetAbsOrigin(vecHelper);
        NetProps.SetPropInt(laser_target_helper_ent, "m_ubInterpolationFrame", frame);

        CreateAoE(player.GetCenter(), 150,
            function (target, deltaVector, distance)
            {
                if (target == player)
                    return;

                if (target.IsPlayer())
                {
                    deltaVector.Norm();
                    deltaVector.x = deltaVector.x * 50
                    deltaVector.y = deltaVector.y * 50
                    deltaVector.z = 50;
                    target.Yeet(deltaVector);
                }

                target.TakeDamageCustom(
                    custom_dmg_eye_lasers,
                    player,
                    player.GetActiveWeapon(),
                    Vector(),
                    player.GetOrigin(),
                    GetMaxHealthCapped(target) * 0.015,
                    0,
                    TF_DMG_CUSTOM_PLASMA);
            });

        CreateAoE(endpos, 100,
            function (target, deltaVector, distance)
            {
                if (target == player)
                    return;

                local damage = clampFloor(2, GetMaxHealthCapped(target) * 0.06 * (1 - distance / 4200.0));
                if (!target.IsPlayer())
                    damage *= 5;

                target.TakeDamageCustom(
                    custom_dmg_eye_lasers,
                    player,
                    player.GetActiveWeapon(),
                    Vector(0, 0, 0),
                    player.GetOrigin(),
                    damage,
                    0,
                    TF_DMG_CUSTOM_PLASMA);
            });
    }

    function GetHUDBarValue()
    {
        return IsWarmup() || IsFiring() ? -1 : GetCooldownTimeStamp();
    }
}