PrecacheModel("models/vip_mobster/weapons/w_moneybag_open.mdl");

class ::MoneyBagProjectile
{
    moneyBagModel = "models/vip_mobster/weapons/w_moneybag_closed.mdl";
    moneyBagModelOpen = "models/vip_mobster/weapons/w_moneybag_open.mdl";
    moneyBagDeploySound = "vip_mobster/weapons/moneybag_deploy.mp3";
    moneyBagParticle = "utaunt_cash_confetti";
    duration = 5.0;

    physprop = null;
    airblastDetector = null;
    player = null;

    constructor(player)
    {
        PrecacheSound(moneyBagDeploySound);
        this.player = player;

        local forwardVelocity = player.EyeAngles().Forward() * 4400 + Vector(0, 0, 200);

        physprop = SpawnEntityFromTable("prop_physics_override", {
            model = moneyBagModel,
            origin = player.EyePosition(),
            spawnflags = "4",
            massscale = 5
        });
        physprop.SetPhysVelocity(forwardVelocity);
        physprop.SetModelScale(1.35, -1);
        physprop.SetPhysAngularVelocity(Vector(-550, -550, 050))
        physprop.ValidateScriptScope();
        physprop.GetScriptScope().script <- this;
        physprop.AcceptInput("DisableShadow", null, null, null);

        RunWithDelay(0.1, function()
        {
            if (this && this.IsValid())
                this.SetPhysVelocity(forwardVelocity);
        }, physprop);

        airblastDetector = CreateByClassname("tf_projectile_cleaver");
        airblastDetector.SetAbsOrigin(physprop.GetCenter());
        SetPropInt(airblastDetector, "m_fEffects", 129);
        SetPropInt(airblastDetector, "m_nNextThinkTick", 0x7FFFFFFF);
        airblastDetector.SetSolid(0);
        airblastDetector.SetMoveType(0, 0);
        airblastDetector.DispatchSpawn();
        airblastDetector.SetModelSimple("models/empty.mdl");
        airblastDetector.SetTeam(player.GetTeam());
        airblastDetector.AcceptInput("SetParent", "!activator", physprop, physprop);

        AddTimer(-1, TickMoneyBagProjectile);
    }

    function TickMoneyBagProjectile()
    {
        if (airblastDetector && GetPropInt(airblastDetector, "m_iDeflected") > 0)
        {
            OnDeflect(GetPropEntity(airblastDetector, "m_hDeflectOwner"));
            airblastDetector.Kill();
            airblastDetector = null;
        }
        local trace = {
            start = physprop.GetOrigin(),
            end = physprop.GetOrigin() - Vector(0, 0, 25),
            ignore = physprop
        };
        TraceLineEx(trace);
        if (!safeget(trace, "enthit", player).IsPlayer())
        {
            MoneyBagFountain(this);
            physprop.Kill();
        }
    }

    function OnDeflect(deflectOwner)
    {
        printl("OnDeflect! "+deflectOwner);
    }
}

class ::MoneyBagFountain
{
    moneyBagOwner = null;
    moneyFountain = null;
    self = null;

    affectedPlayers = null; //{}

    constructor(moneyBagProjectile)
    {
        this.moneyBagOwner = moneyBagProjectile.player;
        local physprop = moneyBagProjectile.physprop;
        local origin = physprop.GetOrigin();

        affectedPlayers = {};

        local trace = {
            start = origin,
            end = origin - Vector(0, 0, 100),
            mask = CONTENTS_HITBOX|CONTENTS_SOLID|CONTENTS_MOVEABLE|CONTENTS_PLAYERCLIP|CONTENTS_WINDOW|CONTENTS_GRATE|CONTENTS_MONSTER,
            ignore = physprop
        };
        TraceLineEx(trace);
        local endpos = safeget(trace, "endpos", null);
        if (endpos)
            origin = endpos;

        local moneyFountain = SpawnEntityFromTable("prop_dynamic",
        {
            origin = origin,
            model = moneyBagProjectile.moneyBagModelOpen,
            teamnum = moneyBagOwner.GetTeam()
        });
        moneyFountain.AcceptInput("DisableShadow", null, null, null);
        moneyFountain.ValidateScriptScope();
        moneyFountain.GetScriptScope().script <- this;
        moneyFountain.SetOwner(moneyBagOwner);
        self = moneyFountain;

        EmitSoundEx({
            sound_name = moneyBagProjectile.moneyBagDeploySound,
            entity = moneyFountain,
            volume = 1,
            sound_level = 90
        });

        for (local x = -175; x <= 200; x += 87)
            for (local y = -175; y <= 200; y += 87)
                EntFireByHandle(SpawnEntityFromTable("info_particle_system",
                {
                    origin = moneyFountain.GetOrigin() + Vector(x, y, 0),
                    effect_name = moneyBagProjectile.moneyBagParticle,
                    start_active = 1
                }), "SetParent", "!activator", -1, moneyFountain, moneyFountain);

        AddTimer(0.2, TickMoneyBagFountain);

        RunWithDelay(moneyBagProjectile.duration, function()
        {
            moneyFountain.Kill();
        })
    }

    function TickMoneyBagFountain()
    {
        local myPos = self.GetOrigin();
        foreach (nearbyPlayer in GetAlivePlayers(self.GetTeam()))
            if ((myPos - nearbyPlayer.GetOrigin()).Length() < 325)
            {
                if (nearbyPlayer in affectedPlayers)
                    affectedPlayers[nearbyPlayer].expireTime = Time() + 1;
                else
                    affectedPlayers[nearbyPlayer] <- MoneyBagEffect(moneyBagOwner, nearbyPlayer);
            }
    }
}

class ::MoneyBagEffect
{
    player = null;
    moneyBagOwner = null;

    expireTime = null;
    antiLoopFlag = false;

    constructor(moneyBagOwner, player)
    {
        this.player = player;
        this.moneyBagOwner = moneyBagOwner;
        expireTime = Time() + 1;

        AddTimer(0.2, function()
        {
            if (!player) //todo
                return TIMER_DELETE;
            player.Heal(player.GetMaxHealth() * 0.03, false);
            if (Time() > expireTime)
            {
                player = null;
                return TIMER_DELETE;
            }
        });

        OnGameEvent("OnTakeDamage", function(victim, params)
        {
            if (player == victim)
                params.damage *= 0.85;
            else if (player == params.attacker)
                params.damage *= 1.15;
        })

        //This is to grant the assist to the VIP. It doesn't actually make people take increased damage.
        OnGameEvent("player_hurt", 0, function(victim, params)
        {
            local attacker = GetPlayerFromParams(params, "attacker");
            if (!attacker || player != attacker || attacker == moneyBagOwner)
                return;
            if (!antiLoopFlag)
            {
                antiLoopFlag = true;
                params.attacker = moneyBagOwner.GetUserID();
                SendGlobalGameEvent("player_hurt", params);
                antiLoopFlag = false;
            }
            //todo failed correct player check
            if (victim.GetHealth() <= 0)
            {
                victim.StunPlayer(0, 0, 2, moneyBagOwner);
                victim.StopSound("TFPlayer.StunImpact");
                attacker.StopSound("TFPlayer.StunImpact");
            }
        })
    }
}